from functools import wraps

from django.core.exceptions import PermissionDenied
from django.utils.module_loading import import_string

from firefence.rules import RuleSet
from firefence.settings import FIREFENCE_SETTINGS


def get_backend_class(import_path=None):
    return import_string(import_path or FIREFENCE_SETTINGS.get('DEFAULT_BACKEND'))


class AbstractFence(object):
    _rules = RuleSet()

    def __init__(self, rules=None):
        self.rules = rules

    @property
    def rules(self):
        return self._rules

    @rules.setter
    def rules(self, value):
        if isinstance(value, RuleSet):
            self._rules = value
        else:
            self._rules = RuleSet(value)

    def allows(self, request):
        return self.rules.allows(request)

    def reject(self, request, *args, **kwargs):
        raise NotImplementedError()  # pragma: no cover

    def protect(self, view_func):
        @wraps(view_func)
        def wrapped_view(request, *args, **kwargs):
            if not self.allows(request):
                return self.reject(request, *args, **kwargs)
            return view_func(request, *args, **kwargs)
        return wrapped_view


class Fence(AbstractFence):
    def reject(self, request, *args, **kwargs):
        raise PermissionDenied



from firefence.backends import get_backend_class
from firefence.settings import FIREFENCE_SETTINGS


class FirefenceMiddleware(object):
    def __init__(self, get_response=None):
        self.get_response = get_response
        super(FirefenceMiddleware, self).__init__()

    def __call__(self, request):
        response = self.process_request(request)
        if not response:
            response = self.get_response(request)
        return response

    def process_request(self, request):
        backend_class = get_backend_class()
        fence = backend_class(FIREFENCE_SETTINGS.get('RULES'))
        if not fence.allows(request):
            return fence.reject(request)



from django.conf import settings


FIREFENCE_SETTINGS = {
    'DEFAULT_BACKEND': 'firefence.backends.Fence',
    'RULES': (),
}

if hasattr(settings, 'FIREFENCE'):
    FIREFENCE_SETTINGS.update(settings.FIREFENCE)



import ipcalc


def get_remote_ip(request):
    xff = request.META.get('HTTP_X_FORWARDED_FOR')
    return xff.split(',')[0] if xff else request.META.get('REMOTE_ADDR')


def get_remote_host(request):
    return request.META.get('HTTP_X_FORWARDED_HOST', request.META.get('REMOTE_HOST'))


def get_server_port(request):
    try:
        return int(request.META.get('HTTP_X_FORWARDED_PORT', request.META.get('SERVER_PORT')))
    except (TypeError, ValueError):
        return None


class Rule(object):
    ALLOW = 'ALLOW'
    DENY = 'DENY'

    _action = None
    _ports = ()
    host = None

    def __init__(self, **kwargs):
        if 'action' not in kwargs:
            raise ValueError('Rules must have an action.')
        self.action = kwargs.get('action')
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')

    def __str__(self):
        string = self.action
        if self.host:
            string += ' from {}'.format(self.host)
        if self.port:
            string += ' to {}'.format(self.port)
        return string or ''

    def __repr__(self):
        return '<Rule: {}>'.format(self)

    @property
    def action(self):
        return self._action

    @action.setter
    def action(self, value):
        if value.upper() not in (self.ALLOW, self.DENY):
            raise ValueError('Action must be ALLOW or DENY.')
        self._action = value.upper()

    @property
    def port(self):
        if self._ports:
            if len(self._ports) == 1:
                return str(self._ports[0])
            else:
                if set(self._ports) == set(range(self._ports[0], self._ports[-1] + 1)):
                    return '{}:{}'.format(self._ports[0], self._ports[-1])
                else:
                    return ', '.join([str(port) for port in self._ports])
        return None

    @port.setter
    def port(self, value):
        if value is None:
            self._ports = ()
        elif isinstance(value, (str, int)):
            ports = str(value).split(',')
            if len(ports) == 1:
                ports = ports[0].split(':', 1)

                if len(ports) == 1:
                    self._ports = (int(ports[0]),)
                else:
                    self._ports = tuple(range(int(ports[0]), int(ports[1]) + 1))
            else:
                self._ports = tuple(sorted([int(port) for port in set(ports)]))
        elif isinstance(value, (list, tuple)):
            self._ports = tuple([int(port) for port in value])
        else:
            raise TypeError(
                'Port must be a string or integer or a list or tuple of strings or integers.')

    def host_matches(self, request):
        remote_ip = get_remote_ip(request)
        remote_host = get_remote_host(request)

        if self.host:
            if remote_host == self.host:
                return True

            try:
                return remote_ip in ipcalc.Network(self.host)
            except ValueError:
                return False

        return True

    def port_matches(self, request):
        server_port = get_server_port(request)
        return server_port in self._ports if self.port else True

    def matches(self, request):
        return self.host_matches(request) and self.port_matches(request)


class RuleSet(object):
    _rules = ()

    def __init__(self, rules=None):
        self.rules = rules

    def __len__(self):
        return len(self.rules)

    def __getitem__(self, index):
        return self.rules[index]

    def __iter__(self):
        return self.rules.__iter__()

    def __str__(self):
        return '\n'.join([str(rule) for rule in self.rules])

    def __repr__(self):
        return repr(self.rules)

    @property
    def rules(self):
        return self._rules

    @rules.setter
    def rules(self, value):
        if value is None:
            self._rules = ()
        else:
            if not isinstance(value, (list, tuple)):
                raise TypeError('Expected a list or tuple of rules.')

            rules = []
            for rule in value:
                if isinstance(rule, Rule):
                    rules.append(rule)
                elif isinstance(rule, dict):
                    rules.append(Rule(**rule))
                else:
                    raise TypeError('Expected a Rule or dict.')
            self._rules = tuple(rules)

    def allows(self, request):
        default = True

        for rule in self.rules:
            if rule.action == Rule.DENY and rule.matches(request):
                return False
            elif rule.action == Rule.ALLOW:
                # Change the default if an ALLOW rule exists
                default = False

                if rule.matches(request):
                    return True

        return default



"""
Like a firewall but smaller.

Protect your views from intruders!
"""
import os

from setuptools import find_packages, setup


DEPENDENCIES = [
    'ipcalc >= 1.99.0',
    'six >= 1.10.0',
]

ROOT = os.path.abspath(os.path.dirname(__file__))

version = __import__('firefence').__version__


setup(
    name='django-firefence',
    version=version,
    url='https://github.com/rehandalal/django-firefence',
    license='Mozilla Public License Version 2.0',
    author='Rehan Dalal',
    author_email='rehan@meet-rehan.com',
    description='A firewall for your Django views.',
    long_description=__doc__,
    packages=find_packages(exclude=['tests', 'docs']),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=DEPENDENCIES,
    py_modules=['firefence'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)



from firefence.backends import get_backend_class
from firefence.settings import FIREFENCE_SETTINGS


def fence_protected(rules=None, backend_class=None):
    backend_class = backend_class or get_backend_class()
    rules = rules or FIREFENCE_SETTINGS.get('RULES')
    fence = backend_class(rules)
    return fence.protect
